from requests.models import Response

from .providers import match as match_provider
from .providers.base import Redirect
from .gen204 import Gen204
from .utils import any_redirect, response_to_str
from .session import session as s


def connect(res: Response) -> bool:
    while True:
        provider = match_provider(res)

        if not provider:
            redirect = any_redirect(res)

            if redirect:
                print(f'Following unknown redirect: {redirect}')
                res = s.get(redirect, allow_redirects=False)
                continue
            else:
                print('----')
                print(response_to_str(res))
                print('----')
                print('No more redirects')
                break

        print(f'Provider: {provider.__class__.__name__}')
        result = provider.run()

        if isinstance(result, Redirect):
            print(f'Following post-auth redirect: {result.url}')
            res = s.get(result.url, allow_redirects=False)
            continue

        if not result.success:
            return False

        break

    print('Checking connection...')
    return Gen204.check().is_connected



__ALL__ = ['AuthWifiRu', 'AuthWifiRuMsk', 'AuthWifiRuSpb']


from furl import furl
from requests import Response
from .base import Provider, Result, Redirect
from ..session import session as s
from ..utils import safeget


class AuthWifiRu(Provider):
    @staticmethod
    def match(response: Response):
        url = furl(response.headers.get('location'))
        return url.host == 'auth.wi-fi.ru' and url.path == '/auth'


class AuthWifiRuMsk(Provider):
    PATHS = {
        'start': '/gapi/auth/start',
        'init': '/gapi/auth/init',
        'check': '/gapi/auth/check'
    }

    @staticmethod
    def match(response: Response):
        url = furl(response.headers.get('location'))
        return url.host == 'auth.wi-fi.ru' and url.path in ['', '/', '/new']

    def run(self) -> Result:
        url = furl(self.response.headers.get('location'))

        segment = url.args.get('segment') or 'metro'
        # client_mac has higher priority
        mac = url.args.get('client_mac') or url.args.get('mac')

        # Follow first redirect
        print('Opening auth page')
        res = s.get(url)
        s.headers['referer'] = str(url)

        # Get auth page
        print('Starting session')
        url.args.clear()
        url.args['segment'] = segment
        if mac:
            url.args['clientMac'] = mac
        url.path = self.PATHS['start']
        res = s.get(url)

        after_auth = safeget(res.json(), 'data', 'segmentParams', 'common',
                             'redirectUrl', 'afterAuth')
        if after_auth:
            print(f'Post-auth redirect: {after_auth}')

        # Send login form
        print('Initializing connection')
        url.path = self.PATHS['init']
        url.args.clear()
        res = s.post(url, data={'mode': 0, 'segment': segment})
        res_data = res.json()
        print(res_data)

        error_code = safeget(res_data, 'auth_error_code', default='')
        if error_code.startswith('err_device_not_identified'):
            print('Error: Device is not registered. Please go to https://wi-fi.ru')
            return Result(False)

        # Checking auth state
        print('Checking connection')
        url.path = self.PATHS['check']
        res = s.get(url)
        res_data = res.json()
        print(res_data)

        if after_auth:
            return Redirect(after_auth)
        else:
            return Result(True)


class AuthWifiRuSpb(AuthWifiRuMsk):
    PATHS = {
        'start': '/spb/gapi/auth/start',
        'init': '/spb/gapi/auth/init',
        'check': '/spb/gapi/auth/check'
    }

    @staticmethod
    def match(response: Response):
        url = furl(response.headers.get('location'))
        return url.host == 'auth.wi-fi.ru' and url.path == '/spb/'



from requests import Response
from typing import Union
from .base import Provider
from .mosmetro import *
from ..utils import all_subclasses


def match(response: Response) -> Union[Provider, None]:
    for subcls in all_subclasses(Provider):
        try:
            if subcls.match(response):
                return subcls(response)
        except Exception:
            pass

    return None



from requests import Response


class Result:
    def __init__(self, success: bool):
        self.success = success


class Redirect(Result):
    def __init__(self, url: str):
        super().__init__(True)
        self.url = url


class Provider:
    def __init__(self, response: Response):
        self.response = response

    def run(self) -> Result:
        raise NotImplementedError()

    @staticmethod
    def match(response: Response):
        raise NotImplementedError()



import sys
from datetime import datetime

from . import __version__
from .gen204 import Gen204
from .providers import Provider
from .utils import all_subclasses
from .connect import connect


def main():
    print(datetime.now())
    print(f'Version: {__version__}')

    print('Loaded providers: ' +
          ', '.join(p.__name__ for p in all_subclasses(Provider)))

    print('Checking connection...')
    res204 = Gen204.check()

    if res204.is_connected:
        print('Already connected')
        sys.exit(0)

    if connect(res204.response):
        print("Connected successfully! :3")
        sys.exit(0)
    else:
        print("Connection failed :(")
        sys.exit(1)


if __name__ == '__main__':
    main()



import re
from bs4.element import Tag
from furl import furl, Path
from requests import Response
from bs4 import BeautifulSoup as BS
from typing import Dict, List, Any, Type, Union


def safeget(data: Dict[Any, Any], *keys, default: Any = None) -> Any:
    """Recursively and safely get a value from nested dictionaries."""
    if not data:
        return None

    res = data

    for key in keys:
        if key in res:
            res = res[key]
        else:
            return default

    return res


def all_subclasses(cls) -> List[Type]:
    """Find all subclasses recursively."""
    res = list()

    for subcls in cls.__subclasses__():
        res.append(subcls)
        res.extend(all_subclasses(subcls))

    return res


def response_to_str(res: Response) -> str:
    return '\n'.join([
        f'{res.status_code} {res.reason}',
        '\n'.join(f'{name}: {value}' for name, value in res.headers.items()),
        '',
        res.text
    ])


def merge_urls(base: Union[str, None], url: Union[str, None]) -> str:
    u1 = furl(base)
    u2 = furl(url)

    if u2.path and not u2.path.isabsolute:
        if u1.path == '':
            raise ValueError('Not enough information to reconstruct URL')

        u2.path = Path(u1.path).add(u2.path)

    if not u2.host:
        if not u1.host:
            raise ValueError('Not enough information to reconstruct URL')

        u2.host = u1.host

    if not u2.scheme:
        if not u1.scheme:
            u1.scheme = 'http'

        u2.scheme = u1.scheme

    return u2.url


PATTERN_META_REDIR = re.compile('^[0-9]+[;,] ?(URL=|url=)?[\'"]?(.*?)[\'"]?$')


def meta_redirect(res: Response) -> Union[str, None]:
    soup = BS(res.content, features="html.parser")
    tag = soup.find('meta', attrs={'http-equiv': re.compile('refresh', re.I)})

    if not tag or not isinstance(tag, Tag):
        return None

    content = tag.attrs['content']
    match = PATTERN_META_REDIR.fullmatch(content)

    if not match:
        return None

    return merge_urls(res.request.url, match.group(2))


def any_redirect(res: Response) -> Union[str, None]:
    """Extract 3xx or meta redirect from Response."""
    if 'location' in res.headers:
        return merge_urls(res.request.url, res.headers['location'])

    return meta_redirect(res)



import random
from typing import List
from requests import Response, RequestException
from requests.exceptions import SSLError
from .session import session as s


class Gen204Res:
    def __init__(self, response: Response = None, false_negative: Response = None):
        self.response = response
        self.false_negative = false_negative

    @property
    def is_connected(self) -> bool:
        if not self.response:
            return False

        return self.response.status_code == 204


class Gen204:
    URL_DEFAULT = [
        'connectivitycheck.gstatic.com/generate_204',
        'www.gstatic.com/generate_204',
        'connectivitycheck.android.com/generate_204',
        'play.googleapis.com/generate_204',
        'clients1.google.com/generate_204'
    ]

    URL_RELIABLE = [
        'www.google.ru/generate_204',
        'www.google.ru/gen_204',
        'google.com/generate_204',
        'gstatic.com/generate_204',
        'maps.google.com/generate_204',
        'mt0.google.com/generate_204',
        'mt1.google.com/generate_204',
        'mt2.google.com/generate_204',
        'mt3.google.com/generate_204',
        'www.google.com/generate_204',
    ]

    @staticmethod
    def request(schema: str, urls: List[str]) -> Response:
        res = None
        last_ex = None

        for base_url in random.choices(urls, k=3):
            url = f'{schema}://{base_url}'

            try:
                res = s.get(url, allow_redirects=False)
                last_ex = None
                print(f'Gen204 | {url} | {res.status_code}')
                break
            except SSLError as ex:
                print(f'Gen204 | {url} | {ex}')
                last_ex = ex
                break
            except RequestException as ex:
                print(f'Gen204 | {url} | {ex}')
                last_ex = ex

        if last_ex:
            raise last_ex

        return res

    @staticmethod
    def check() -> Gen204Res:
        """Returns gen204 response and false negative (if exists)."""
        # Unreliable HTTP check (needs to be verified by HTTPS)
        try:
            unrel = Gen204.request("http", Gen204.URL_DEFAULT)
        except RequestException:
            # network is most probably unreachable
            return Gen204Res()

        # Reliable HTTPS check
        try:
            rel_https = Gen204.request("https", Gen204.URL_RELIABLE)
        except RequestException:
            rel_https = None

        if unrel.status_code == 204:
            if not rel_https or rel_https.status_code != 204:
                # Reliable HTTP check
                try:
                    rel_http = Gen204.request("http", Gen204.URL_RELIABLE)
                except RequestException:
                    rel_http = None

                if rel_http and rel_http.status_code != 204:
                    return Gen204Res(rel_http)  # false positive
            else:
                return Gen204Res(rel_https)  # confirmed positive
        else:
            if not rel_https:
                return Gen204Res(unrel)  # confirmed negative
            elif rel_https.status_code == 204:
                return Gen204Res(rel_https, unrel)  # false negative

        print('Gen204 | Unexpected state')
        return Gen204Res()


if __name__ == '__main__':
    res_204 = Gen204.check()
    print(f'Connected: {res_204.is_connected}')
    print(f'False negative: {res_204.false_negative is not None}')



#!/usr/bin/env python

from setuptools import setup, find_packages


with open('README.md', 'r') as fi:
    long_description = fi.read()


setup(
    name='mosmetro',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],

    author='Dmitry Karikh',
    author_email='the.dr.hax@gmail.com',

    description='Скрипт для автоматической авторизации в сетях московского (и не только) общественного транспорта',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mosmetro-android/mosmetro-python',
    license='GNU GPLv3',

    install_requires=[
        'requests',
        'furl',
        'user_agent',
        'beautifulsoup4'
    ],

    packages=find_packages(),
    include_package_data=True,

    entry_points='''
        [console_scripts]
        mosmetro=mosmetro.__main__:main
    ''',

    classifiers=[
        'Environment :: Console',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Internet',
        'Topic :: Utilities'
    ]
)
